# -*- coding: utf-8 -*-
from __future__ import print_function, unicode_literals

from pypeg2 import *

pypeg2_parse = parse  # rename it,for avoiding name conflict
tag = re.compile(r"\d+")
nomeaning = blank, maybe_some(comment_sh), blank
fullname = re.compile(r"[\w+\.]*\w+")


class MainKey(str):
    grammar = "(", optional(word), ")"


class Decimal(int):
    grammar = "(", optional(tag), ")"


class TypeName(object):
    grammar = flag("is_arr", "*"), attr("fullname", fullname)


class Filed(List):
    grammar = (
        attr("filed", word),
        attr("tag", tag),
        ":",
        attr("typename", TypeName),
        optional(Decimal),
        optional(MainKey),
        nomeaning,
        endl,
    )


class Struct(List):
    pass


class Type(List):
    pass


Struct.grammar = "{", nomeaning, attr("fileds", maybe_some([Filed, Type])), "}"
Type.grammar = nomeaning, ".", name(), attr("struct", Struct), nomeaning


class Sub_pro_type(Keyword):
    grammar = Enum(K("request"), K("response"))


class Subprotocol(List):
    grammar = (
        attr("subpro_type", Sub_pro_type),
        attr("pro_filed", [TypeName, Struct]),
        nomeaning,
    )


class Protocol(List):
    grammar = (
        nomeaning,
        attr("name", word),
        attr("tag", tag),
        "{",
        nomeaning,
        attr("fileds", maybe_some(Subprotocol)),
        "}",
        nomeaning,
    )


class Sproto(List):
    grammar = attr("items", maybe_some([Type, Protocol]))


# ====================================================================

builtin_types = {
    "integer": 0,
    "boolean": 1,
    "string": 2,
    "double": 3,
    "binary": 2,
}  # add double and binary

import re as rawre


def checktype(types, ptype, t):
    if t in builtin_types:
        return t

    fullname = "%s.%s" % (ptype, t)
    if fullname in types:
        return fullname
    else:
        sobj = rawre.search("(.+)\..+$", ptype)
        if sobj:
            return checktype(types, sobj.group(1), t)
        elif t in types:
            return t


def flattypename(r):
    for typename, t in r["type"].items():
        for _, f in enumerate(t):
            ftype = f["typename"]
            fullname = checktype(r["type"], typename, ftype)
            assert fullname != None, "Undefined type %s in type %s" % (ftype, typename)
            f["typename"] = fullname
            if "array" in f and "key" in f:
                pass


class Convert:
    group = {}
    type_dict = {}
    protocol_dict = {}
    protocol_tags = {}  # just for easiliy check

    @staticmethod
    def parse(text, name="=text"):
        Convert.group = {}
        Convert.type_dict = {}
        Convert.protocol_dict = {}
        Convert.protocol_tags = {}

        obj = pypeg2_parse(text, Sproto)
        for i in obj.items:
            if hasattr(i, "tag"):
                Convert.convert_protocol(i)
            else:
                Convert.convert_type(i)

        Convert.group["type"] = Convert.type_dict
        Convert.group["protocol"] = Convert.protocol_dict

        # import json
        # print(json.dumps(Convert.group, indent=4))
        return Convert.group

    @staticmethod
    def convert_type(obj, parent_name=""):  # todo add mainkey
        if parent_name != "":
            obj.name = parent_name + "." + obj.name
        type_name = obj.name
        if type_name in Convert.type_dict.keys():
            print("Error:redifine %s\n" % (type_name))
            return False
        Convert.type_dict[type_name] = Convert.convert_struct(obj.struct, type_name)

    @staticmethod
    def convert_struct(obj, name=""):  # todo 你 加入decimal
        struct = []
        for filed in obj.fileds:
            if type(filed) == Filed:
                filed_typename = filed.typename.fullname
                filed_type = Convert.get_typename(filed_typename)
                filed_info = {}
                filed_info["name"] = filed.filed
                filed_info["tag"] = int(filed.tag)
                filed_info["array"] = filed.typename.is_arr
                filed_info["typename"] = filed_typename
                filed_info["type"] = filed_type
                if len(filed) > 0:
                    if filed_typename == "integer":
                        filed_info["decimal"] = filed[0]
                    else:
                        filed_info["key"] = filed[0]  # todo 解决冲突

                struct.append(filed_info)
            elif type(filed) == Type:
                Convert.convert_type(filed, name)
        return struct

    @staticmethod
    def convert_protocol(obj):
        if obj.name in Convert.protocol_dict.keys():
            print("Error:redifine protocol %s \n" % (obj.name))
            return
        if obj.tag in Convert.protocol_tags.keys():
            print("Error:redifine protocol tags %d \n" % (obj.tag))
            return
        protocol = {}
        protocol["tag"] = int(obj.tag)
        protocol["name"] = obj.name
        for fi in obj.fileds:
            if type(fi.pro_filed) == TypeName:
                assert fi.pro_filed.is_arr == False, "syntax error at %s.%s" % (
                    obj.name,
                    fi.subpro_type,
                )
                newtype_name = str("".join(fi.pro_filed.fullname))
                protocol[fi.subpro_type] = newtype_name
            elif type(fi.pro_filed) == Struct:
                newtype_name = obj.name + "." + fi.subpro_type
                Convert.type_dict[newtype_name] = Convert.convert_struct(
                    fi.pro_filed, newtype_name
                )
                protocol[fi.subpro_type] = newtype_name

        Convert.protocol_dict[obj.name] = protocol
        Convert.protocol_tags[obj.tag] = True

    @staticmethod
    def get_typename(name):
        if name in builtin_types:
            return "builtin"
        else:
            return "UserDefine"


# ===============================================================
# export functions
# ===============================================================
__all__ = ["parse", "parse_list", "builtin_types"]


def parse(text, name="=text", check=True):
    build = Convert.parse(text, name)
    if check:
        flattypename(build)
    return build


def parse_list(sproto_list):
    build = {"protocol": {}, "type": {}}
    for v in sproto_list:
        ast = Convert.parse(v[0], v[1])

        # merge type
        for stname, stype in ast["type"].iteritems():
            assert stname not in build["type"], "redifine type %s in %s" % (
                stname,
                v[1],
            )
            build["type"][stname] = stype
        # merge protocol
        for spname, sp in ast["protocol"].iteritems():
            assert (
                spname not in build["protocol"]
            ), "redifine protocol name %s in %s" % (spname, v[1])
            for proto in build["protocol"]:
                assert (
                    sp["tag"] != build["protocol"][proto]["tag"]
                ), "redifine protocol tag %d in %s with %s" % (sp["tag"], proto, spname)
            build["protocol"][spname] = sp

    flattypename(build)
    # checkprotocol(build)
    return build



# cython: language_level=3
cimport cython
from cpython.bool cimport PyBool_FromLong
from cpython.bytes cimport (PyBytes_AsStringAndSize, PyBytes_Check,
                            PyBytes_FromStringAndSize)
from cpython.dict cimport (PyDict_Check, PyDict_GetItemString, PyDict_New,
                           PyDict_SetItem, PyDict_SetItemString)
from cpython.exc cimport PyErr_Occurred, PyErr_Print
from cpython.float cimport PyFloat_AsDouble, PyFloat_Check, PyFloat_FromDouble
from cpython.list cimport PyList_Append, PyList_Check, PyList_New, PyList_Size
from cpython.long cimport (PyLong_AsLong, PyLong_AsLongLong, PyLong_Check,
                           PyLong_FromLong, PyLong_FromLongLong)
from cpython.mem cimport PyMem_Free, PyMem_Malloc, PyMem_Realloc
from cpython.object cimport PyObject
from cpython.ref cimport Py_TYPE
from cpython.unicode cimport PyUnicode_DecodeUTF8
from libc.stdint cimport int32_t, int64_t, uint8_t
from libc.string cimport memcpy

from pysproto cimport sproto


cdef extern from "Python.h":
    char* PyUnicode_AsUTF8AndSize(object data, Py_ssize_t* l)


class SprotoError(Exception):
    pass

cdef struct encode_ud:
    PyObject *data  # dict
    PyObject *values

# cdef struct decode_ud:
#     PyObject* data # type: dict
#     PyObject* key
#     int deep
#     int mainindex_tag

cdef struct decode_ud:
    int mainindex
    PyObject *data  # dict
    PyObject *map_key

cdef bytes _ensure_bytes(s):
    if s is None:
        return s
    elif isinstance(s, unicode):
        return (<unicode>s).encode()
    elif isinstance(s, bytearray):
        s = bytes(s)
    elif not isinstance(s, bytes):
        raise ValueError("expected string, got %s" % type(s))
    return <bytes>s

cdef int encode(const sproto.sproto_arg *args) except * with gil:
    cdef:
        encode_ud *self = <encode_ud*>args.ud
        const char * tagname = args.tagname
        int type = args.type
        int index = args.index
        int length = args.length
        object data
        int count
        long long i, vh

        char* ptr
        Py_ssize_t l

        encode_ud sub
    cdef PyObject* tmp =  PyDict_GetItemString(<object>self.data, tagname)
    if tmp == NULL:
        if index > 0:
            return sproto.SPROTO_CB_NOARRAY
        return sproto.SPROTO_CB_NIL
    data = <object>tmp
    if index > 0:
        if not PyList_Check(data):
            if PyDict_Check(data):
                count = 0
                for key, value in data.items():
                    count +=1
                    if index == count:
                        data = value
                        break
                if index > count:
                    return sproto.SPROTO_CB_NIL
            else:
                try:
                    raise SprotoError("Expected List or Dict for tagname:%s" % tagname)
                finally:
                    return sproto.SPROTO_CB_ERROR
        else:
            if index > <int>PyList_Size(data):
                # printf("data is finish, index:%d, len:%d\n", index, len)
                return sproto.SPROTO_CB_NIL
            data = data[index - 1]

    if type == sproto.SPROTO_TINTEGER:
        if PyLong_Check(data) or (args.extra and PyFloat_Check(data)):
            if args.extra:
                i = <long long>(PyFloat_AsDouble(data) * args.extra + 0.5)
                # printf("input data:%lld\n", i)
            else:
                i = PyLong_AsLongLong(data)
            vh = i >> 31
            if vh == 0 or vh == -1:
                (<int32_t *>args.value)[0] = <int32_t>i
                return 4
            else:
                (<int64_t *>args.value)[0] = <int64_t>i
                return 8
        else:
            try:
                raise SprotoError("type mismatch, tag:%s, expected int or long, got:%s\n", tagname, Py_TYPE(data).tp_name)
            finally:
                return -1
    elif type == sproto.SPROTO_TBOOLEAN: # https://github.com/spin6lock/python-sproto/blob/py3_test/src/pysproto/python_sproto.c#L111
        if isinstance(data, bool):
            if data:
                (<int *> args.value)[0] = 1
            else:
                (<int *> args.value)[0] = 0
        return 4
    elif type == sproto.SPROTO_TSTRING: # L120
        if args.extra == sproto.SPROTO_TSTRING_BINARY:   # binary
            if not PyBytes_Check(data):
                try:
                    raise SprotoError("type mismatch, tag:%s, expected bytes, got:%s\n", tagname, Py_TYPE(data).tp_name)
                finally:
                    return sproto.SPROTO_CB_ERROR
            PyBytes_AsStringAndSize(data, &ptr, &l)
        else:
            if not isinstance(data, str):
                try:
                    raise SprotoError("type mismatch, tag:%s, expected unicode, got:%s\n", tagname, Py_TYPE(data).tp_name)
                finally:
                    return sproto.SPROTO_CB_ERROR
            ptr = <char*>PyUnicode_AsUTF8AndSize(data, &l)
        if <int>l>length:
            return sproto.SPROTO_CB_ERROR
        memcpy(args.value, ptr, <size_t> l)
        return <int>l
    elif type == sproto.SPROTO_TSTRUCT: # L152
        sub.data = <PyObject*>data
        with nogil:
            r = sproto.sproto_encode(args.subtype, args.value, length, encode, &sub)
        if r<0:
            return sproto.SPROTO_CB_ERROR
        return r
    else:
        return 0

# cdef int decode(const sproto.sproto_arg *args) except * with gil: # except * with gil
#     cdef decode_ud *ud = <decode_ud *> args.ud
#     self_d = <dict> ud.data
#     # todo: need check deep?
#     if args.index != 0:
#         if args.tagname not in self_d:
#             if args.mainindex >= 0:
#                 c = {}
#             else:
#                 c = []
#             self_d[args.tagname] = c
#         else:
#             c = self_d[args.tagname]
#         if args.index < 0:
#             return 0
#
#     ret = None
#     cdef decode_ud sub
#     if args.type == sproto.SPROTO_TINTEGER:
#         if args.extra:
#             ret = (<int64_t *> args.value)[0]
#             ret = <double> ret / args.extra
#         else:
#             ret = (<int64_t *> args.value)[0]
#     elif args.type == sproto.SPROTO_TBOOLEAN:
#         ret = True if (<int64_t *> args.value)[0] > 0 else False
#     elif args.type == sproto.SPROTO_TSTRING:
#         ret = (<char *> args.value)[:args.length]
#     elif args.type == sproto.SPROTO_TSTRUCT:
#         sub.deep = ud.deep + 1
#         sub_d = {}
#         sub.data = <PyObject *> sub_d
#         if args.mainindex >= 0:
#             sub.mainindex_tag = args.mainindex
#             r = sproto.sproto_decode(args.subtype, args.value, args.length, decode, &sub)
#             if r < 0:
#                 return sproto.SPROTO_CB_ERROR
#             if r != args.length:
#                 return r
#             if sub.key == NULL:
#                 raise SprotoError("can't find mainindex (tag_or_name=%d) in [%s]a" % (args.mainindex, args.tagname))
#             c[<object> (sub.key)] = sub_d
#             return 0
#         else:
#             sub.mainindex_tag = -1
#             r = sproto.sproto_decode(args.subtype, args.value, args.length, decode, &sub)
#             if r < 0:
#                 return sproto.SPROTO_CB_ERROR
#             if r != args.length:
#                 return r
#             ret = sub_d
#     else:
#         raise SprotoError("Decode error, got invalid type %d" % args.type)
#
#     if args.index > 0:
#         c.append(ret)
#     else:
#         if ud.mainindex_tag == args.tagid:
#             ud.key = <PyObject *> ret
#         self_d[args.tagname] = ret
#     return 0


cdef int decode(const sproto.sproto_arg *args) except * with gil: # except * with gil
    cdef:
        decode_ud *self = <decode_ud *>args.ud
        const char *tagname = args.tagname
        int tagid = args.tagid
        int type = args.type
        int index = args.index
        int mainindex = args.mainindex
        Py_ssize_t length = args.length

        PyObject * obj=self.data
        object data = None # to return

        # long long i
        int64_t tmp
        # double result
        decode_ud sub
    if index != 0:
        obj = PyDict_GetItemString(<object>self.data, tagname)
        if obj == NULL:
            if mainindex >= 0:
                d = PyDict_New()
                obj = <PyObject*>d
            else:
                l = PyList_New(0)
                obj = <PyObject*>l
            PyDict_SetItemString(<object>self.data, tagname, <object>obj)
            if index < 0:
                return 0
    if type == sproto.SPROTO_TINTEGER:
        # args.extra = 0 length = 8
        if args.extra:
            tmp = (<int64_t*>args.value)[0]
            # result = <double>tmp / args.extra
            data = PyFloat_FromDouble(<double>tmp / args.extra)
        elif length == 4:
            data = PyLong_FromLong((<int32_t*>args.value)[0])
        elif length == 8:
            data = PyLong_FromLongLong((<int64_t *> args.value)[0]) # data = 1
        else:
            try:
                raise SprotoError("unexpected integer length: %d" % length)
            finally:
                return sproto.SPROTO_CB_ERROR
    elif type == sproto.SPROTO_TBOOLEAN:
        data = PyBool_FromLong((<int *>args.value)[0])
    elif type == sproto.SPROTO_TSTRING:
        if args.extra == sproto.SPROTO_TSTRING_STRING:
            data = PyUnicode_DecodeUTF8(<char*>args.value, length, "can not decode utf8-encode unicode string")
        else:
            data = PyBytes_FromStringAndSize(<char*>args.value, length)
    elif type == sproto.SPROTO_TSTRUCT:
        d = PyDict_New()
        sub.data = <PyObject*>d
        if mainindex >= 0:
            sub.mainindex = args.mainindex
            with nogil:
                r = sproto.sproto_decode(args.subtype, args.value, <int>length, decode, &sub)
            if r<0:
                return sproto.SPROTO_CB_ERROR
            if r!=length:
                return r
            PyDict_SetItem(<object>obj, <object>sub.map_key, <object>sub.data)
        else:
            sub.mainindex = -1
            data = <object>sub.data
            with nogil:
                r = sproto.sproto_decode(args.subtype, args.value, <int>length, decode, &sub)
            if r<0:
                return sproto.SPROTO_CB_ERROR
            if r!=length:
                return r
    if data:
        if PyList_Check(<object>obj):
            PyList_Append(<object>obj , data)
        else:
            PyDict_SetItemString(<object>self.data, tagname, data)
        # Py_XDECREF(<PyObject*>data) catch you! damn refcnt!
        if self.mainindex == tagid:
            self.map_key = <PyObject*>data
    else:
        if self.mainindex == tagid:
            try:
                raise SprotoError("map key type not support")
            finally:
                return sproto.SPROTO_CB_ERROR
    return 0


@cython.freelist(8)
@cython.no_gc
@cython.final
cdef class SprotoType:
    """Wrapper around struct sproto_type"""
    cdef sproto.sproto_type_t *st

    @staticmethod
    cdef inline SprotoType from_ptr(sproto.sproto_type_t *st):
        cdef SprotoType self = SprotoType.__new__(SprotoType)
        self.st = st
        return self

    @property
    def name(self):
        return (<bytes>sproto.sproto_name(self.st)).decode()

    cpdef inline dict decode(self, const uint8_t[::1] buffer):
        assert self.st != NULL
        cdef:
            dict d = {}
            decode_ud ud
            int r
        ud.data = <PyObject*>d
        with nogil:
            r = sproto.sproto_decode(self.st, <void*>&buffer[0], <int>buffer.shape[0], decode, &ud)
        if PyErr_Occurred():
            PyErr_Print()
            raise SprotoError("decode error")
        if r < 0:
            raise SprotoError("decode error")
        return d

    cpdef inline int encode_into(self, dict data, uint8_t[::1] buffer) except -1:
        """
        encode data into buffer
        :param data: 
        :param buffer: 
        :return: buffer updated
        """
        assert self.st != NULL
        cdef:
            int ret
            encode_ud ud = encode_ud(<PyObject*>data, NULL)
        with nogil:
            ret = sproto.sproto_encode(self.st, <void*>&buffer[0], <int>buffer.shape[0], encode, &ud)
        if ret < 0:
            raise SprotoError("buffer is too small")
        return ret

    cpdef inline bytes encode(self, dict data):
        assert self.st != NULL
        cdef:
            int ret
            encode_ud ud = encode_ud(<PyObject *> data, NULL)
            size_t prealloc = 1024
        cdef uint8_t * buf = <uint8_t *>PyMem_Malloc(prealloc)
        if buf == NULL:
            raise MemoryError
        try:
            while True:
                with nogil:
                    ret = sproto.sproto_encode(self.st, buf, <int>prealloc, encode, &ud)
                if PyErr_Occurred():
                    PyErr_Print()
                    raise SprotoError("encode error")
                if ret < 0:
                    prealloc = prealloc * 2
                    buf = <uint8_t *> PyMem_Realloc(buf, prealloc)
                    if buf == NULL:
                        raise MemoryError
                else:
                    return <bytes>buf[:ret]
        finally:
            PyMem_Free(buf)

@cython.freelist(8)
@cython.no_gc
@cython.final
cdef class Sproto:
    """Wrapper around struct sproto"""
    cdef sproto.sproto * sp

    def __cinit__(self, const uint8_t[::1] buffer):
        self.sp = sproto.sproto_create(<void*>&buffer[0], buffer.shape[0])
        if self.sp is NULL:
            raise MemoryError

    def __dealloc__(self):
        if self.sp != NULL:
            sproto.sproto_release(self.sp)
            self.sp = NULL

    def dump(self):
        assert self.sp != NULL
        with nogil:
            sproto.sproto_dump(self.sp)

    cpdef inline SprotoType querytype(self, type_name):
        assert self.sp != NULL
        type_name = _ensure_bytes(type_name)
        cdef sproto.sproto_type_t *st
        cdef char* type_name_c = <char*>type_name
        with nogil:
            st = sproto.sproto_type(self.sp, type_name_c)
        if st:
            return SprotoType.from_ptr(st)

    cpdef inline object protocol(self, object tag_or_name):
        assert self.sp != NULL
        cdef:
            const char* name
            int tag = 0
            sproto.sproto_type_t * request
            sproto.sproto_type_t * response
            object ret1, ret2, ret3  # firtst ret params
        if isinstance(tag_or_name, int):
            tag = <int>PyLong_AsLong(tag_or_name)
            with nogil:
                name = sproto.sproto_protoname(self.sp, tag)
            if name == NULL:
                return None
            ret1  = (<bytes>name).decode()
        elif isinstance(tag_or_name, (str, bytes)):
            bt = _ensure_bytes(tag_or_name)
            name = <const char*>bt
            with nogil:
                tag = sproto.sproto_prototag(self.sp, name)
            if tag < 0:
                return None
            ret1  = PyLong_FromLong(<int>tag)
        with nogil:
            request = sproto.sproto_protoquery(self.sp, tag, sproto.SPROTO_REQUEST)
        if request == NULL:
            ret2 = None
        else:
            ret2 = SprotoType.from_ptr(request)
        with nogil:
            response = sproto.sproto_protoquery(self.sp, tag, sproto.SPROTO_RESPONSE)
        if response == NULL:
            ret3 = None
        else:
            ret3 = SprotoType.from_ptr(response)
        return (ret1, ret2, ret3)

    cpdef inline int sproto_protoresponse(self, int proto):
        cdef int ret
        with nogil:
            ret = sproto.sproto_protoresponse(self.sp, proto)
        return ret

cpdef inline int pack_into(const uint8_t[::1] inp, uint8_t[::1] out) except -1:
    cdef:
        size_t sz = <size_t>inp.shape[0]
        size_t maxsz = (sz + 2047) / 2048 * 2 + sz + 2
        int ret
    if <size_t>out.shape[0] < maxsz:
        raise SprotoError("output buffer is too small")
    with nogil:
        ret =  sproto.sproto_pack(<void*>&inp[0], <int>inp.shape[0], <void*>&out[0], <int>maxsz)
    if <size_t>ret > maxsz:
        raise SprotoError("packing error, return size = %d" % ret)
    return ret


cpdef inline bytes pack(const uint8_t[::1] inp):
    cdef:
        size_t sz = <size_t> inp.shape[0]
        size_t maxsz = (sz + 2047) / 2048 * 2 + sz + 2
        int ret
    cdef void* out = PyMem_Malloc(maxsz)
    if out == NULL:
        raise MemoryError
    with nogil:
        ret = sproto.sproto_pack(<void *> &inp[0], <int> inp.shape[0], out, <int>maxsz)
    if <size_t>ret > maxsz:
        raise SprotoError("packing error, return size = %d" % ret)
    bt = <bytes>((<uint8_t*>out)[:ret])
    PyMem_Free(out)
    return bt


cpdef inline int unpack_into(const uint8_t[::1] inp, uint8_t[::1] out) except -1:
    cdef int ret
    with nogil:
        ret =  sproto.sproto_unpack(<void*>&inp[0], <int>inp.shape[0], <void*>&out[0], <int>out.shape[0])
    if ret < 0:
        raise SprotoError("Invalid unpack stream")
    return ret


cpdef inline bytes unpack(const uint8_t[::1] inp):
    cdef void * out = PyMem_Malloc(<size_t>sproto.ENCODE_BUFFERSIZE)
    if out == NULL:
        raise MemoryError
    cdef int osz = sproto.ENCODE_BUFFERSIZE
    cdef int ret
    with nogil:
        ret = sproto.sproto_unpack(<void *> &inp[0], <int> inp.shape[0], out, osz)
    if ret < 0:
        raise SprotoError("Invalid unpack stream")
    if ret > osz:
        out = PyMem_Realloc(out, <size_t>ret)
        if out == NULL:
            raise MemoryError
        osz = ret
        with nogil:
            ret = sproto.sproto_unpack(<void *> &inp[0], <int> inp.shape[0], out, osz)
        if ret < 0:
            raise SprotoError("Invalid unpack stream")
    bt = <bytes>((<uint8_t*>out)[:ret])
    PyMem_Free(out)
    return bt



# -*- coding: utf-8 -*-
import os
import re
from collections import defaultdict

from Cython.Build import cythonize
from setuptools import Extension, find_packages, setup
from setuptools.command.build_ext import build_ext

BUILD_ARGS = defaultdict(lambda: ["-O3", "-g0"])

for compiler, args in [
    ("msvc", ["/EHsc", "/DHUNSPELL_STATIC", "/Oi", "/O2", "/Ot"]),
    ("gcc", ["-O3", "-g0"]),
]:
    BUILD_ARGS[compiler] = args


class build_ext_compiler_check(build_ext):
    def build_extensions(self):
        compiler = self.compiler.compiler_type
        args = BUILD_ARGS[compiler]
        for ext in self.extensions:
            ext.extra_compile_args = args
        super().build_extensions()


extensions = [
    Extension(
        "pysproto._sproto",
        ["pysproto/_sproto.pyx", f"sproto/sproto.c"],
        include_dirs=[f"./sproto"],
        library_dirs=[f"./sproto"],
    ),
]


def get_dis():
    with open("README.markdown", "r", encoding="utf-8") as f:
        return f.read()


def get_version() -> str:
    path = os.path.join(
        os.path.abspath(os.path.dirname(__file__)), "pysproto", "__init__.py"
    )
    with open(path, "r", encoding="utf-8") as f:
        data = f.read()
    result = re.findall(r"(?<=__version__ = \")\S+(?=\")", data)
    return result[0]


packages = find_packages(exclude=("test", "tests.*", "test*"))


def main():
    version: str = get_version()
    dis = get_dis()
    setup(
        name="sproto",
        version=version,
        url="https://github.com/synodriver/sproto",
        packages=packages,
        keywords=["sproto"],
        description="sproto encode and decode",
        long_description_content_type="text/markdown",
        long_description=dis,
        author="synodriver",
        author_email="diguohuangjiajinweijun@gmail.com",
        python_requires=">=3.6",
        setup_requires=["cython"],
        install_requires=["pypeg2"],
        license="BSD",
        classifiers=[
            "Development Status :: 4 - Beta",
            "Operating System :: OS Independent",
            "License :: OSI Approved :: BSD License",
            "Programming Language :: C",
            "Programming Language :: Cython",
            "Programming Language :: Python",
            "Programming Language :: Python :: 3.6",
            "Programming Language :: Python :: 3.7",
            "Programming Language :: Python :: 3.8",
            "Programming Language :: Python :: 3.9",
            "Programming Language :: Python :: 3.10",
            "Programming Language :: Python :: Implementation :: CPython",
        ],
        include_package_data=True,
        zip_safe=False,
        cmdclass={"build_ext": build_ext_compiler_check},
        ext_modules=cythonize(
            extensions,
            compiler_directives={
                "cdivision": True,
                "embedsignature": True,
                "boundscheck": False,
                "wraparound": False,
            },
        ),
    )


if __name__ == "__main__":
    main()



# coding:utf-8
import argparse
import os
import struct
import sys
from io import BytesIO
from typing import Union

import pysproto.sprotoparser as sprotoparser


def ensure_bytes(data) -> bytes:
    if isinstance(data, str):
        return data.encode()
    return bytes(data)


def packbytes(s: Union[bytes, str]) -> bytes:
    s = ensure_bytes(s)
    return struct.pack("<I%ds" % len(s), len(s), s)


def packvalue(v: int) -> bytes:
    v = (v + 1) * 2
    return struct.pack("<H", v)


def packfield(f):
    strtbl = BytesIO()  # todo L289
    if f["array"]:
        if f["key"]:  # if has no "key" already set to f["key"] = None
            if f.get("map", None):
                strtbl.write(b"\7\0")
            else:
                strtbl.write(b"\6\0")
        else:
            strtbl.write(b"\5\0")
    else:
        strtbl.write(b"\4\0")
    strtbl.write(b"\0\0")
    if f["builtin"] != None:
        strtbl.write(packvalue(f["builtin"]))
        if f.get("extra", None):
            strtbl.write(packvalue(f["extra"]))
        else:
            strtbl.write(b"\1\0")
        strtbl.write(packvalue(f["tag"]))
    else:
        strtbl.write(b"\1\0")
        strtbl.write(packvalue(f["type"]))
        strtbl.write(packvalue(f["tag"]))
    if f["array"]:
        strtbl.write(packvalue(1))
        if f["key"]:
            strtbl.write(packvalue(f["key"]))
            if f.get("map", None):
                strtbl.write(packvalue(f["map"]))
    strtbl.write(packbytes(f["name"]))
    return packbytes(strtbl.getvalue())


def packtype(name, t, alltypes):
    fields = []
    tmp = {}
    for f in t:  # type: dict
        tmp["array"] = f["array"]
        tmp["name"] = f["name"]
        tmp["tag"] = f["tag"]
        tmp["extra"] = f.get("decimal", None)

        tname = f["typename"]
        tmp["builtin"] = sprotoparser.builtin_types.get(tname, None)
        subtype = None

        if tname == "binary":
            tmp["extra"] = 1
        if tmp["builtin"] == None:
            assert alltypes[tname], "type %s not exists" % tname
            subtype = alltypes[tname]
            tmp["type"] = subtype["id"]
        else:
            tmp["type"] = None
        if f.get("key", None) is not None:  # todo Line 352
            assert f["array"]
            if f["key"] == "":
                tmp["map"] = 1
                c = 0
                min_t = sys.maxsize
                for n, t in enumerate(subtype["fields"]):
                    c += 1
                    if t["tag"] < min_t:
                        min_t = t["tag"]
                        f["key"] = n
                assert c == 2, (
                    "Invalid map definition: %s, must only have two fields"
                    % tmp["name"]
                )
            stfield = subtype["fields"].get(f.get("key", None), None)
            if not stfield or not stfield.get("buildin", None):
                raise AssertionError("Invalid map index :" + f["key"])
            tmp["key"] = stfield.get("tag", None)

            # tmp["key"] = subtype["fields"][f["key"]["name"]]
            # assert tmp["key"], "Invalid map index %d" % f["key"]["name"]
        else:
            tmp["key"] = None
        fields.append(packfield(tmp))
    data = BytesIO()
    if not fields:
        data.write(b"\1\0\0\0")
        data.write(packbytes(name))
        # data = [b"\1\0", b"\0\0", packbytes(name)]
    else:
        data.write(b"\2\0\0\0\0\0")
        data.write(packbytes(name))
        data.write(packbytes(b"".join(fields)))
        # data = [b"\2\0", b"\0\0", b"\0\0", packbytes(name), packbytes(b"".join(fields))]
    return packbytes(data.getvalue())


def packproto(name, p, alltypes) -> bytes:
    if "request" in p:
        request = alltypes[p["request"]]
        assert request != None, "Protocol %s request types not found" % (
            name,
            p["request"],
        )
        request = request["id"]

    tmp = BytesIO()
    tmp.write(b"\4\0\0\0")
    tmp.write(packvalue(p["tag"]))

    # tmp = ["\4\0", "\0\0", packvalue(p["tag"])]
    if "request" not in p and "response" not in p:
        tmp.getbuffer()[:2] = b"\2\0"
        # tmp[0] = "\2\0"
    else:
        if "request" in p:
            tmp.write(packvalue(alltypes[p["request"]]["id"]))
        else:
            tmp.write(b"\1\0")
        if "response" in p:
            tmp.write(packvalue(alltypes[p["response"]]["id"]))
        else:
            tmp.getbuffer()[:2] = b"\3\0"
    tmp.write(packbytes(name))
    return packbytes(tmp.getvalue())


def packgroup(t, p) -> bytes:
    """

    :param t: Type
    :param p: Protocol
    :return:
    """
    if not t:
        assert p
        return b"\0\0"
    tp = None
    alltypes = {}
    alltype_names = []
    for name in t:
        alltype_names.append(name)
    alltype_names.sort()
    for idx, name in enumerate(alltype_names):
        fields = {}
        for type_fields in t[name]:
            if (
                type_fields["typename"] in sprotoparser.builtin_types
            ):  # todo add key too nested
                fields[type_fields["name"]] = type_fields["tag"]
        alltypes[name] = {"id": idx, "fields": fields}

    tt = BytesIO()
    for name in alltype_names:
        tt.write(
            packtype(name, t[name], alltypes)
        )  # alltypes["Person"]["fields"]["key"]

    tt = packbytes(tt.getvalue())
    if p:
        tmp = []
        for name, tbl in p.iteritems():
            tmp.append(tbl)
            tbl["name"] = name
        tmp = sorted(tmp, key=lambda k: k["tag"])
        tp = BytesIO()
        for tbl in tmp:
            tp.write(packproto(tbl["name"], tbl, alltypes))
        tp = packbytes(tp.getvalue())
    result = BytesIO()
    if tp == None:
        result.write(b"\1\0\0\0")
        result.write(tt)
        # result = [b"\1\0", b"\0\0", tt]
    else:
        result.write(b"\2\0\0\0\0\0")
        result.write(tt)
        result.write(tp)
        # result = [b"\2\0", b"\0\0", b"\0\0", tt, tp]
    return result.getvalue()


def encodeall(r) -> bytes:  # todo 彻底核对lua的对应这个名字的函数
    return packgroup(r["type"], r["protocol"])


def parse_ast(ast) -> bytes:
    return encodeall(ast)


def dump(build, outfile):
    data = parse_ast(build)
    if isinstance(outfile, str):
        f = open(outfile, "wb")
        f.write(data)
        f.close()
    else:
        outfile.write(data)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("-d", "--directory", dest="src_dir", help="sproto source files")
    parser.add_argument("-f", "--file", dest="src_file", help="sproto single file")
    parser.add_argument(
        "-o",
        "--out",
        dest="outfile",
        default="sproto.spb",
        help="specific dump binary file",
    )
    parser.add_argument(
        "-v", "--verbose", dest="verbose", action="store_true", help="show more info"
    )
    args = parser.parse_args()

    build = None
    if args.src_file:
        text = open(args.src_file, encoding="utf-8").read()
        build = sprotoparser.parse(text, os.path.basename(args.src_file))
    else:
        sproto_list = []
        for f in os.listdir(args.src_dir):
            file_path = os.path.join(args.src_dir, f)
            if os.path.isfile(file_path) and f.endswith(".sproto"):
                text = open(file_path, encoding="utf-8").read()
                sproto_list.append((text, f))

        build = sprotoparser.parse_list(sproto_list)

    if args.verbose == True:
        import json

        print(json.dumps(build, indent=4))
    dump(build, args.outfile)
